"""Utility functions."""
from __future__ import absolute_import

from .download import download
from .filesystem import makedirs
from . import bbox
from .bbox import bbox_iou
from .block import recursive_visit, set_lr_mult, freeze_bn
from . import viz
from . import random
from . import metrics

from .lr_scheduler import PolyLRScheduler
from .metrics.voc_segmentation import batch_pix_accuracy, batch_intersection_union
from . import parallel

from .plot_history import TrainingHistory
